"""
    Defines library global variables
"""
from pymatriz.enums import MessageType
from pymatriz.parser import MarketDataMessageParser, BookMessageParser, ConnectionStatusMessageParser

config = {
    "ssl": True,
    "url": "https://mtzdma.primary.ventures/",
    "ws_url": "wss://mtzdma.primary.ventures/",
    "message_parsers": {
        MessageType.MarketData: MarketDataMessageParser,
        MessageType.Book: BookMessageParser,
        MessageType.Portfolio: None,
        MessageType.Order: None,
        MessageType.OrderEvent: None,
        MessageType.News: None,
        MessageType.ExternalMarketData: None,
        MessageType.ConnectionStatus: ConnectionStatusMessageParser,
        MessageType.Clock: None,
        MessageType.CancelReject: None,
        MessageType.Control: None,
        MessageType.Ignore: None,
    }
}


from pymatriz.enums import FieldType, MessageType
from pymatriz.exceptions import ParsingException


class MessageParser:
    @staticmethod
    def parse(message):
        raise NotImplementedError


class MarketDataMessageParser(MessageParser):

    _MD_FIELDS = 17

    @staticmethod
    def parse(message):
        fields = message.split("|")

        if len(fields) < MarketDataMessageParser._MD_FIELDS:
            raise ParsingException(f"Invalid tick length: {len(fields)}.")

        return {
            FieldType.Type.value: MessageType.MarketData.value,
            FieldType.SymbolId.value: fields[0],
            FieldType.Seq.value: fields[1],
            FieldType.Offers.value: fields[2] if fields[2] != "" else None,
            FieldType.Asz.value: fields[3] if fields[3] != "" else None,
            FieldType.Bids.value: fields[4] if fields[4] != "" else None,
            FieldType.Bsz.value: fields[5] if fields[5] != "" else None,
            FieldType.Last.value: fields[6] if fields[6] != "" else None,
            FieldType.OpeningPrice.value: fields[7] if fields[7] != "" else None,
            FieldType.ClosingPrice.value: fields[8] if fields[8] != "" else None,
            FieldType.LowPrice.value: fields[9] if fields[9] != "" else None,
            FieldType.HighPrice.value: fields[10] if fields[10] != "" else None,
            FieldType.NominalVolume.value: fields[11] if fields[11] != "" else None,
            FieldType.TradeVolume.value: fields[12] if fields[12] != "" else None,
            FieldType.Time.value: fields[13],
            FieldType.SettlementPrice.value: fields[14] if fields[14] != "" else None,
            FieldType.OpenInterest.value: fields[15] if fields[15] != "" else None,
            FieldType.Refp.value: fields[16] if fields[16] != "" else None
        }


class BookMessageParser(MessageParser):

    _BOOK_FIELDS = 5

    @staticmethod
    def parse(message):
        message_fields = message.split("!")

        if len(message_fields) < BookMessageParser._BOOK_FIELDS:
            raise ParsingException(f"Invalid tick length: {len(message_fields)}.")

        sid = message_fields[0]
        seq = message_fields[1]
        books = []

        for content in message_fields[2:5]:
            fields = content.split("|")

            books.append({
                FieldType.Bsz.value: fields[0] if fields[0] != "" else None,
                FieldType.Bids.value: fields[1] if fields[1] != "" else None,
                FieldType.Offers.value: fields[2] if fields[2] != "" else None,
                FieldType.Asz.value: fields[3] if fields[3] != "" else None,
            })

        return {
            FieldType.Type.value: MessageType.Book.value,
            FieldType.SymbolId.value: sid,
            FieldType.Seq.value: seq,
            FieldType.Books.value: books
        }


class ConnectionStatusMessageParser(MessageParser):
    @staticmethod
    def parse(message):
        return {
            FieldType.Type.value: MessageType.ConnectionStatus.value,
            FieldType.Status.value: "connected" == message
        }



from pymatriz import messages
from pymatriz.enums import FieldType, MessageType, Market, MarketDataEntry
import pandas as pd


class Client:

    def __init__(self, config, settings=None):
        self.config = config
        self.settings = settings

        # Handlers for incoming messages
        self.market_data_handlers = []
        self.order_report_handlers = []
        self.error_handlers = []
        self.exception_handler = None

    def set_settings(self, settings):
        self.settings = settings

    def add_market_data_handler(self, handler):
        """ Adds a new Market Data handler to the handlers list.

        :param handler: function that is going to be call when a new Market Data Message is received.
        :type handler: callable.
        """
        if handler not in self.market_data_handlers:
            self.market_data_handlers.append(handler)

    def remove_market_data_handler(self, handler):
        """ Removes the Market Data handler from the handler list.

        :param handler: function to be removed from the handler list.
        :type handler: callable.
        """
        if handler in self.market_data_handlers:
            self.market_data_handlers.remove(handler)

    def add_order_report_handler(self, handler):
        """ Adds a new Order Report handler to the handlers list.

        :param handler: function that is going to be call when a new Order Report Message is received.
        :type handler: callable.
        """
        if handler not in self.order_report_handlers:
            self.order_report_handlers.append(handler)

    def remove_order_report_handler(self, handler):
        """ Removes the Order Report handler from the handler list.

        :param handler: function to be removed from the handler list.
        :type handler: callable.
        """
        if handler in self.order_report_handlers:
            self.order_report_handlers.remove(handler)

    def add_error_handler(self, handler):
        """ Adds a new Error handler to the handlers list.

        :param handler: function that is going to be call when a new Error Message is received.
        :type handler: callable.
        """
        if handler not in self.error_handlers:
            self.error_handlers.append(handler)

    def remove_error_handler(self, handler):
        """ Removes the Error handler from the handler list.

        :param handler: function to be removed from the handler list.
        :type handler: callable.
        """
        if handler in self.error_handlers:
            self.error_handlers.remove(handler)

    def set_exception_handler(self, handler):
        """ Sets the Exception Handler.

        :param handler: function called when Exception is raised.
        :type handler: callable.
        """
        self.exception_handler = handler

    def on_message(self, message):
        """ Called when a new message is received through the connection.

        :param message: message received.
        :type message: str
        """
        try:
            if message != "pong":
                if "\n" in message:
                    for line in message.split("\n"):
                        if ";" in line:
                            msg = self.parse_message(line[0], line[2:])
                            if msg:
                                self.handle_message(msg)
                else:
                    msg = self.parse_message(message[0], message[2:])
                    if msg:
                        self.handle_message(msg)

        except Exception as e:
            self.on_exception(e)

    def handle_message(self, msg):
        try:

            # Checks if it is an error message
            if FieldType.Type.value in msg:
                # extract the message type.
                msg_type = msg[FieldType.Type.value]

                # Checks message type and call the correct handlers
                if msg_type == MessageType.MarketData.value:
                    for handler in self.market_data_handlers:
                        df = pd.DataFrame(msg, index=[0])
                        df = df.drop([FieldType.Type.value], axis=1)
                        # df.set_index([FieldType.SymbolId.value, FieldType.Time.Type])
                        handler(df)
                elif msg_type == MessageType.Book.value:
                    for handler in self.order_report_handlers:
                        df = pd.DataFrame(msg, index=[0])
                        df = df.drop([FieldType.Type.value], axis=1)
                        handler(df)
                elif msg_type == MessageType.ConnectionStatus.value:
                    # self.connected = msg[FieldType.Status]
                    pass
                else:
                    msg_type_not_supported = "Websocket: Message Type not Supported. Message: {msg}"
                    for handler in self.error_handlers:
                        handler(msg_type_not_supported.format(msg=msg))
            else:
                msg_not_supported = "Websocket: Message Supported. Message: {msg}"
                for handler in self.error_handlers:
                    handler(msg_not_supported.format(msg=msg))

        except Exception as e:
            self.on_exception(e)

    def parse_message(self, msg_type, payload):
        parser = self.config["message_parsers"].get(MessageType(msg_type))

        if parser:
            return parser.parse(payload)

        return False

    def on_exception(self, exception):
        """Called when an exception occurred within the client.

        :param exception: exception raised.
        :type exception: exception object
        """
        self.exception_handler(exception)

    def _url(self, path):
        raise NotImplementedError


class ApiClient(Client):
    def __init__(self, config, settings=None):
        super().__init__(config, settings)

    def build_instruments(self, tickers, **kwargs):
        market = kwargs["market"] if "markets" in kwargs else Market.MERVAL
        terms = kwargs["terms"] if "terms" in kwargs else [MarketDataEntry.TERM_48HS]

        has_to_append_market = market.value not in tickers[0]
        has_to_append_terms = MarketDataEntry.TERM_CI.value not in tickers[0] and \
                              MarketDataEntry.TERM_24HS.value not in tickers[0] and \
                              MarketDataEntry.TERM_48HS.value not in tickers[0]

        if has_to_append_market:
            if has_to_append_terms:
                instruments = [
                    messages.DOUBLE_QUOTES.format(item=
                                                  messages.INSTRUMENT_TERM.format(ticker=ticker,
                                                                                  market=market.value,
                                                                                  term=term.value))
                    for term in terms
                    for ticker in tickers
                ]
            else:
                instruments = [
                    messages.DOUBLE_QUOTES.format(item=
                                                  messages.INSTRUMENT.format(ticker=ticker,
                                                                             market=market.value))
                    for ticker in tickers
                ]
        else:
            instruments = [messages.DOUBLE_QUOTES.format(item=ticker) for ticker in tickers]

        return instruments

    def get_messages(self, message):
        messages = []
        try:
            if message != "pong":
                if "\n" in message:
                    for line in message.split("\n"):
                        if ";" in line:
                            messages.append(self.parse_message(line[0], line[2:]))
                else:
                    messages.append(self.parse_message(message[0], message[2:]))

            return messages
        except Exception as e:
            self.on_exception(e)

    def get_market_data(self):
        raise NotImplementedError

    def get_intraday_history(self, tickers, **kwargs):
        raise NotImplementedError

    def get_daily_history(self, tickers, **kwargs):
        raise NotImplementedError

    def get_all_instruments(self):
        raise NotImplementedError



import csv
from collections import defaultdict
from datetime import datetime, timedelta

import pandas as pd
from io import StringIO

import requests
import logging
from http.client import HTTPConnection  # py3
from lxml import html
import re
import base64
import json

from pymatriz import urls, messages
from pymatriz.client_interface import ApiClient
from pymatriz.enums import Market, FieldType, HistoryFieldType
from pymatriz.exceptions import ApiException


class RestClient(ApiClient):
    """ Rest client that implements call to Matriz Rest API
    """

    def __init__(self, config):
        """Initialization of the Client.
        :param config: Any
        """
        super().__init__(config)

        # self.log = logging.getLogger('urllib3')
        # self.log.setLevel(logging.ERROR)
        #
        # # logging from urllib3 to console
        # ch = logging.StreamHandler()
        # ch.setLevel(logging.ERROR)
        # self.log.addHandler(ch)
        #
        # # print statements from `http.client.HTTPConnection` to console/stdout
        # HTTPConnection.debuglevel = 0

        self.session = requests.Session()

    def update_token(self):
        """Get Matriz settings object, which contains useful data like web token, websocket connection id, RIMA web token.
        :return: Any
        """
        username = self.config["username"]
        password = self.config["password"]

        # Matriz Web Platform Auth0 client ID
        client_id = "LlRkDft0IVkhGLXoknIaOtvlcc7eVAfX"
        auth0_api_id = "matriz.auth0.com"
        auth0_client = "eyJuYW1lIjoibG9jay5qcyIsInZlcnNpb24iOiIxMS41LjIiLCJsaWJfdmVyc2lvbiI6IjkuNC4yIn0="
        realm = "dma3-prod-db"
        credential_type = "http://auth0.com/oauth/grant-type/password-realm"

        matriz_url = "https://mtzdma.primary.ventures"
        redirect_uri = "https://mtzdma.primary.ventures/auth/auth0/callback"

        # Get Auth0 Login ticket
        payload = json.dumps({
            "client_id": client_id,
            "username": username,
            "password": password,
            "realm": realm,
            "credential_type": credential_type
        })

        headers = {
            'authority': auth0_api_id,
            'auth0-client': auth0_client,
            'origin': matriz_url,
            'referer': matriz_url,
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'cross-site',
            'Content-Type': 'application/json',
        }

        response = self.session.request("POST", urls.authenticate, headers=headers, data=payload)

        # Get Auth0 Access token
        url = urls.authorize.format(client_id=client_id,
                                    response_type="token",
                                    response_mode="form_post",
                                    redirect_uri=redirect_uri,
                                    connection=realm,
                                    realm=realm,
                                    scope="openid profile email",
                                    login_ticket=response.json()["login_ticket"],
                                    auth0_client=auth0_client)

        payload = {}
        headers = {
            'authority': auth0_api_id,
            'upgrade-insecure-requests': '1',
            'sec-fetch-site': 'cross-site',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-user': '?1',
            'sec-fetch-dest': 'document',
            'referer': matriz_url,
        }

        response = self.session.request("GET", url, headers=headers, data=payload)
        root = html.fromstring(response.text.encode('utf8'))

        # Get Matriz Platform settings object
        payload = dict(map(lambda i: (i.name, i.value), root.xpath("//form/input")))

        response = self.session.post(root.xpath("//form")[0].action, data=payload)

        settings = json.loads(base64.b64decode(re.search("var settings = '(.*)'", response.text).groups()[0]))

        self.settings = settings

        return settings

    def get_market_data(self):
        response = self.api_request(urls.market_data)

        df = pd.DataFrame(self.get_messages(response.text))

        df = df.drop([FieldType.Type.value], axis=1)

        return df

    def get_intraday_history(self, tickers, **kwargs):
        return self._get_history_call(urls.historical_series_intraday, tickers, **kwargs)

    def get_daily_history(self, tickers, **kwargs):
        return self._get_history_call(urls.historical_series_daily, tickers, **kwargs)

    def _get_history_call(self, url_template, tickers, **kwargs):
        responses = []

        start_date = datetime.fromordinal(kwargs["start_date"].toordinal()).isoformat("T") + "Z" \
            if "start_date" in kwargs else (datetime.utcnow() - timedelta(days=1)).isoformat("T") + "Z"
        end_date = datetime.fromordinal(kwargs["end_date"].toordinal()).isoformat("T") + "Z" \
            if "end_date" in kwargs else (datetime.utcnow() + timedelta(days=1)).isoformat("T") + "Z"

        instruments = self.build_instruments(tickers, **kwargs)

        for instrument in instruments:
            url = url_template.format(instrument=instrument.strip('"'), start_date=start_date, end_date=end_date)
            response = self.api_request(url)

            content = response.content.decode('utf-8')
            r = pd.read_csv(StringIO(content), sep=",")
            r.insert(0, HistoryFieldType.SymbolId.value, instrument.strip('"'))
            responses.extend(r.to_dict(orient='records'))

        df = pd.DataFrame(responses)

        if len(df) > 0:
            index = []

            if HistoryFieldType.SymbolId.value in df:
                index.append(HistoryFieldType.SymbolId.value)

            if HistoryFieldType.Time.value in df:
                index.append(HistoryFieldType.Time.value)

            df.set_index(index, inplace=True)

        return df

    def get_all_instruments(self):
        """Make a request to the API and get a list of all available instruments.

        For more detailed information go to: https://apihub.primary.com.ar/assets/docs/Primary-API.pdf

        :return: A list of valid instruments returned by the API.
        :rtype: dict of JSON response.
        """
        response = self.api_request(urls.instruments)

        return pd.DataFrame(json.loads(response.text))

    def market_data_subscription(self, tickers, **kwargs):
        """ Creates and sends new Market Data Subscription Message through the connection.

        :param tickers: List of the tickers to subscribe.
        :type tickers: list of str
        :param market: Market id associated to the tickers.
        :type market: Market (Enum).
        :param terms: List of terms that want to be received.
        Example: [MarketDataEntry.TERM_CI, MarketDataEntry.TERM_24HS]
        :type terms: List of MarketDataEntry (Enum).
        """

        # Iterates through the tickers list and creates a new list of Instrument String using the INSTRUMENT Template.
        # Then create a comma separated string with the instruments in the list.
        instruments = self.build_instruments(tickers, **kwargs)

        instruments_string = ",".join(instruments)

        # Creates a Market Data Subscription Message using the Template.
        message = messages.MARKET_DATA_SUBSCRIPTION.format(instruments=instruments_string)

        # Send the message through the connection.
        response = self.api_post(urls.ws_subscribe.format(token=self.settings["token"], connection_id=self.settings["connId"]), message)

    def api_request(self, path, retry=True):
        """ Make a GET request to the API.
        :param path: path to the API resource.
        :type path: str
        :param retry: (optional) True: update the token and resend the request if the response code is 401.
        False: raise an exception if the response code is 401.
        :type retry: str
        :return: response of the API.
        :rtype: Response
        """
        headers = {'authorization': f'Bearer {self.settings["token"]}'}
        response = self.session.get(self._url(path),
                                    headers=headers,
                                    verify=self.config["ssl"] if "ssl" in self.config else True,
                                    proxies=self.config["proxies"] if "proxies" in self.config else None)

        # Checks if the response code is 401 (Unauthorized)
        if response.status_code == 401:
            if retry:
                self.update_token()
                self.api_request(path, False)
            else:
                raise ApiException("Authentication Fails.")
        elif response.status_code < 200 or response.status_code > 299:
            raise ApiException(f"Failure requesting {path}. Response ({response.status_code}): {response.text}")

        return response

    def api_post(self, path, data, retry=True):
        """ Make a POST request to the API.
        :param path: path to the API resource.
        :type path: str
        :param retry: (optional) True: update the token and resend the request if the response code is 401.
        False: raise an exception if the response code is 401.
        :type retry: str
        :return: response of the API.
        :rtype: Response
        """
        headers = {'authorization': f'Bearer {self.settings["token"]}', 'Origin': self.config["url"], 'content-type': 'application/json', 'x-csrf-token': self.settings["csrfToken"]}
        response = self.session.post(self._url(path),
                                     data,
                                     headers=headers,
                                     verify=self.config["ssl"] if "ssl" in self.config else True,
                                     proxies=self.config["proxies"] if "proxies" in self.config else None)

        # Checks if the response code is 401 (Unauthorized)
        if response.status_code == 401:
            if retry:
                self.update_token()
                self.api_post(path, data, False)
            else:
                raise ApiException("Authentication Fails.")
        elif response.status_code < 200 or response.status_code > 299:
            raise ApiException(f"Failure requesting {path}. Response ({response.status_code}): {response.text}")

        return response

    def _url(self, path):
        """ Helper function that concatenate the path to the API url.
        :param path: path to the API resource.
        :type path: str
        :return: URL to be call.
        :rtype: str
        """
        return self.config["url"] + path



from pymatriz.client_interface import ApiClient
from pymatriz.enums import Market
from pymatriz.rest_client import RestClient
from pymatriz.websocket_client import WebSocketClient
from pymatriz.globals import config


class MatrizAPIClient(ApiClient):
    """ Matriz API client
    """

    def __init__(self, **kwargs):
        """Initialization of the Client.
        :param config: Any
         """
        if "username" not in kwargs:
            raise Exception("Missing required parameter 'username'")

        if "password" not in kwargs:
            raise Exception("Missing required parameter 'username'")

        config["username"] = kwargs["username"]
        config["password"] = kwargs["password"]
        config["account"] = kwargs["account"] if "account" in kwargs else None

        super().__init__(config)

        self.rest_client = RestClient(self.config)
        self.ws_client = WebSocketClient(self.config)

    def connect(self):
        # Get auth token and other useful settings
        settings = self.rest_client.update_token()
        self.set_settings(settings)

        self.ws_client.connect()

    def market_data_subscription(self, tickers, **kwargs):
        self.rest_client.market_data_subscription(tickers, **kwargs)

    def get_daily_history(self, tickers, **kwargs):
        self.market_data_subscription(tickers, **kwargs)

        return self.rest_client.get_daily_history(tickers, **kwargs)

    def get_intraday_history(self, tickers, **kwargs):
        self.market_data_subscription(tickers, **kwargs)

        return self.rest_client.get_intraday_history(tickers, **kwargs)

    def get_market_data(self):
        self.market_data_subscription(self.settings["favorites"])
        return self.rest_client.get_market_data()

    def set_exception_handler(self, handler):
        super().set_exception_handler(handler)
        self.rest_client.set_exception_handler(handler)
        self.ws_client.set_exception_handler(handler)

    def add_market_data_handler(self, handler):
        super().add_market_data_handler(handler)
        self.rest_client.add_market_data_handler(handler)
        self.ws_client.add_market_data_handler(handler)

    def remove_market_data_handler(self, handler):
        super().remove_market_data_handler(handler)
        self.rest_client.remove_market_data_handler(handler)
        self.ws_client.remove_market_data_handler(handler)

    def add_error_handler(self, handler):
        super().add_error_handler(handler)
        self.rest_client.add_error_handler(handler)
        self.ws_client.add_error_handler(handler)

    def remove_error_handler(self, handler):
        super().remove_error_handler(handler)
        self.rest_client.remove_error_handler(handler)
        self.ws_client.remove_error_handler(handler)

    def get_all_instruments(self):
        return self.rest_client.get_all_instruments()

    def add_order_report_handler(self, handler):
        raise NotImplemented

    def remove_order_report_handler(self, handler):
        raise NotImplemented

    def set_settings(self, settings):
        super().set_settings(settings)
        self.rest_client.set_settings(settings)
        self.ws_client.set_settings(settings)

    def close(self):
        self.ws_client.close_connection()



import time, threading


class set_interval:
    def __init__(self, interval, action):
        self.interval = interval
        self.action = action
        self.stop_event = threading.Event()
        thread = threading.Thread(target=self.__set_interval)
        thread.start()

    def __set_interval(self):
        next_time = time.time() + self.interval
        while not self.stop_event.wait(next_time - time.time()):
            next_time += self.interval
            self.action()

    def cancel(self):
        self.stop_event.set()



from enum import Enum


class Market(Enum):
    """Market ID associated to the instruments.
    ROFEX: ROFEX Exchange.
    """
    ROFEX = 'ROFX'
    MERVAL = 'MERV - XMEV'


class MarketDataEntry(Enum):
    TERM_CI = 'CI'
    TERM_24HS = '24hs'
    TERM_48HS = '48hs'


class MessageType(Enum):
    MarketData = 'm'
    Book = 'b'
    Portfolio = 'p'
    Order = 'o'
    OrderEvent = 'e'
    News = 'n'
    ExternalMarketData = 'x'
    ConnectionStatus = 'c'
    Clock = 'k'
    CancelReject = 'r'
    Control = '$'
    Ignore = '_'


class HistoryFieldType(Enum):
    SymbolId = 'sid'
    Time = 'time'
    OpeningPrice = 'open'
    ClosingPrice = 'close'
    LowPrice = 'low'
    HighPrice = 'high'
    TradeVolume = 'volume'


class FieldType(Enum):
    SymbolId = 'sid'
    Seq = 'seq'
    Offers = 'ask'
    Asz = 'asz'
    Bids = 'bid'
    Bsz = 'bsz'
    Last = 'lst'
    OpeningPrice = 'opn'
    ClosingPrice = 'cls'
    LowPrice = 'low'
    HighPrice = 'hgh'
    NominalVolume = 'von'
    TradeVolume = 'vol'
    Time = 'hor'
    SettlementPrice = 'set'
    OpenInterest = 'oin'
    Refp = 'refp'

    Type = 'type'
    Books = 'books'
    Status = 'status'

    # get_all_instruments
    Aon = 'aon'  # false
    Cfi = 'cfi'  # "ESXXXX"
    Currency = 'cur'  # "ARS"
    Decimals = 'dec'  # 3
    Description = 'des'  # "Rigolleau"
    Exp = 'exp'  # null
    Fix = 'fix'  # "MERV - XMEV - RIGO - 24hs"
    Fun = 'fun'  # "Rigolleau Merval"
    Grp = 'grp'  # null
    Hid = 'hid'  # true
    Iso = 'iso'  # 0
    Isu = 'isu'  # 0
    Mdet = 'mdet'  # "0124578Cxw"
    Market = 'mkt'  # "bm"
    Mpt = 'mpt'  # 0.1
    Mst = 'mst'  # 1
    Ord = 'ord'  # 1
    Prd = 'prd'  # null
    Pxq = 'pxq'  # 1
    Sdc = 'sdc'  # 0
    Segment = 'seg'  # "bm_MERV"
    Ser = 'ser'  # "RIGO"
    Srt = 'srt'  # "bm:RIGO:c:24hs"
    Str = 'str'  # null
    Symbol = 'sym'  # "RIGO (24hs)"
    TypeAcronym = 'typ'  # null
    Und = 'und'  # "MERV - XMEV - RIGO - 48hs"
    Vtm = 'vtm'  # null
    Vty = 'vty'  # null
